from petisco.domain.base_object import BaseObject
from petisco.events.event import Event, Events


class AggregateRoot(BaseObject):
    def __init__(self):
        self.domain_events: Events = []

    def record(self, event: Event):
        self.domain_events.append(event)

    def clear_domain_events(self):
        self.domain_events = []

    def pull_domain_events(self) -> Events:
        return self.domain_events

    def pull_first_domain_event(self) -> Event:
        return self.domain_events[0] if len(self.domain_events) > 0 else None

    def pull_last_domain_event(self) -> Event:
        return self.domain_events[-1] if len(self.domain_events) > 0 else None



class Token:
    def __init__(self, client_id: str, user_id: str, token_type: str):
        self.client_id = client_id
        self.user_id = user_id
        self.token_type = token_type

        if not self.user_id or self.user_id == "null":
            self.user_id = None

    def __repr__(self):
        return f"[Token: [client_id: {self.client_id} | user_id: {self.user_id} | token_type: {self.token_type}]]"

    @staticmethod
    def from_token_payload(token_payload: dict):
        return Token(
            client_id=token_payload.get("cli"),
            user_id=token_payload.get("sub"),
            token_type=token_payload.get("typ"),
        )



from typing import Callable, Dict

from dataclasses import dataclass

from petisco.application.config.get_funtion_from_string import get_function_from_string
from petisco.application.config.raise_petisco_config_error import (
    raise_petisco_config_exception,
)


@dataclass
class ConfigCronJob:
    seconds: float
    handler_key: str
    kdict: Dict = None

    @staticmethod
    def from_dict(kdict):
        seconds = kdict.get("seconds")
        handler_key = kdict.get("handler")
        if not handler_key:
            raise TypeError(f"ConfigEventsSubscriber: handler is a required parameter")

        return ConfigCronJob(seconds, handler_key, kdict)

    def get_handler(self) -> Callable:
        # Use this function to avoid circular dependency accessing Petisco objects before its configuration
        handler = (
            get_function_from_string(self.handler_key)
            .handle(
                on_failure=raise_petisco_config_exception,
                failure_args=(self.kdict, "cron:*:handler"),
            )
            .unwrap()
        )
        return handler



from typing import Optional, Dict

from dataclasses import dataclass

from petisco.application.config.cron.config_cron_job import ConfigCronJob


@dataclass
class ConfigCron:
    jobs: Optional[Dict[str, ConfigCronJob]] = None

    @staticmethod
    def from_dict(kdict):
        if not kdict or not isinstance(kdict, dict):
            return ConfigCron({})

        jobs = {}
        for key, config_job in kdict.items():
            jobs[key] = ConfigCronJob.from_dict(config_job)

        return ConfigCron(jobs)



from typing import Optional, Callable

from dataclasses import dataclass

from petisco.application.config.get_funtion_from_string import get_function_from_string
from petisco.application.config.raise_petisco_config_error import (
    raise_petisco_config_exception,
)
from petisco.events.publisher.infrastructure.not_implemented_event_publisher import (
    NotImplementedEventPublisher,
)


@dataclass
class ConfigEventsPublisher:
    provider: Optional[Callable] = lambda: NotImplementedEventPublisher()

    @staticmethod
    def from_dict(kdict):

        provider = kdict.get("provider")
        if not provider:
            raise TypeError(
                f"ConfigEventsPublisher: {provider} is a required parameter"
            )

        provider = (
            get_function_from_string(provider)
            .handle(
                on_failure=raise_petisco_config_exception,
                failure_args=(kdict, "events:publisher:provider"),
            )
            .unwrap()
        )

        return ConfigEventsPublisher(provider=provider)



import threading
from typing import Dict

from petisco.events.rabbitmq.create_exchange_and_bind_queue import (
    create_exchange_and_bind_queue,
    create_dead_letter_exchange_and_bind_queue,
)
from petisco.events.rabbitmq.get_event_binding_key import get_event_binding_key
from petisco.events.rabbitmq.rabbitmq_connector import RabbitMQConnector
from petisco.events.subscriber.domain.config_event_subscriber import (
    ConfigEventSubscriber,
)
from petisco.events.subscriber.domain.interface_event_subscriber import IEventSubscriber


class RabbitMQEventSubscriber(IEventSubscriber):
    def __init__(
        self,
        connector: RabbitMQConnector,
        subscribers: Dict[str, ConfigEventSubscriber] = None,
        connection_name: str = "subscriber",
    ):
        self._is_subscribed = False
        if not connector:
            raise TypeError(f"RabbitMQEventSubscriber: Invalid Given RabbitMQConnector")
        self.connector = connector
        self.connection = None
        self.connection_name = connection_name
        super().__init__(subscribers)

    def _connect(self):
        self.connection = self.connector.get_connection(self.connection_name)

    def subscribe_all(self):
        self._connect()
        if self.subscribers:
            # Run the worker into a thread
            self._thread = threading.Thread(target=self._subscribe)
            self._thread.start()
            self._is_subscribed = True

    def _subscribe(self):
        self._channels = {}

        # Create channels
        for name in self.subscribers.keys():
            self._channels[name] = self.connection.channel()
            # This uses the basic.qos protocol method to tell RabbitMQ not to give more than one message to a worker at a
            # time. Or, in other words, don't dispatch a new message to a worker until it has processed and acknowledged
            # the previous one. Instead, it will dispatch it to the next worker that is not still busy.
            self._channels[name].basic_qos(prefetch_count=1)

        # Subscription
        for name, subscriber_config in self.subscribers.items():
            self._setup_exchanges_and_queues(subscriber_config)
            queue = (
                subscriber_config.topic
                if not subscriber_config.dead_letter
                else f"dl-{subscriber_config.topic}"
            )
            self._channels[name].basic_consume(
                queue=queue, on_message_callback=subscriber_config.get_handler()
            )

        # Start consuming channels
        for name in self.subscribers.keys():
            self._channels[name].start_consuming()

    def _setup_exchanges_and_queues(self, subscriber_config: ConfigEventSubscriber):
        exchange = subscriber_config.service
        queue = subscriber_config.topic
        binding_key = get_event_binding_key(
            subscriber_config.organization, subscriber_config.service
        )

        create_dead_letter_exchange_and_bind_queue(
            connection=self.connection,
            exchange=exchange,
            queue=queue,
            binding_key=binding_key,
        )
        create_exchange_and_bind_queue(
            connection=self.connection,
            exchange=exchange,
            queue=queue,
            binding_key=binding_key,
            dead_letter=True,
        )

    def unsubscribe_all(self):
        def kill():
            for name in self.subscribers.keys():
                self._channels[name].stop_consuming()

        if self._is_subscribed:
            self.connection.call_later(0, kill)
            self._thread.join()

    def info(self) -> Dict:
        is_open = False
        if self.connection:
            is_open = self.connection.is_open
        return {
            "name": self.__class__.__name__,
            "connection.is_open": is_open,
            "is_subscribed": self._is_subscribed,
        }



import atexit

from apscheduler.schedulers import SchedulerAlreadyRunningError
from apscheduler.schedulers.blocking import BlockingScheduler

from petisco.application.config.cron.config_cron import ConfigCron
from petisco.cron.domain.interface_cron_executor import CronExecutor
from apscheduler.schedulers.background import BackgroundScheduler


class APSchedulerCronExecutor(CronExecutor):
    def __init__(self, scheduler: BlockingScheduler = BackgroundScheduler()):
        self.scheduler = scheduler

    def start(self, config_cron: ConfigCron):
        for config_cron_job in config_cron.jobs.values():
            self.scheduler.add_job(
                func=config_cron_job.get_handler(),
                trigger="interval",
                seconds=config_cron_job.seconds,
            )

        try:
            self.scheduler.start()
            atexit.register(lambda: self.scheduler.shutdown())
        except SchedulerAlreadyRunningError:
            pass

    def stop(self):
        self.scheduler.shutdown()



from typing import Dict

from meiga import Result, Error, Failure, Success

from petisco.application.petisco import Petisco
from petisco.use_case.use_case import UseCase
from petisco.modules.healthcheck.domain.persistence_error import PersistenceError


class HealthcheckProvider(UseCase):
    def execute(self, petisco: Petisco) -> Result[Dict, Error]:

        healthcheck = {"app_name": petisco.app_name, "app_version": petisco.app_version}

        if petisco.persistence_configured:
            from petisco.persistence.sqlalchemy.sqlalchemy_session_scope import (
                session_scope,
            )

            with session_scope() as session:
                try:
                    session.execute("SELECT 1")
                except Exception as e:
                    return Failure(PersistenceError(str(e)))
            healthcheck["persistence_available"] = True

        return Success(healthcheck)



import sys
import importlib
import traceback
from typing import Callable

from meiga import Result, Error, Success, isFailure


def get_function_from_string(function_string: str) -> Result[Callable, Error]:
    if not function_string:
        return Success(None)
    try:
        mod_name, func_name = function_string.rsplit(".", 1)
        mod = importlib.import_module(mod_name)
        func = getattr(mod, func_name)
        return Success(func)
    except Exception as err:
        traceback.print_exc()
        print(err, file=sys.stderr)
        return isFailure



from typing import Optional, Dict, Callable

from dataclasses import dataclass

from petisco.application.config.get_funtion_from_string import get_function_from_string
from petisco.application.config.raise_petisco_config_error import (
    raise_petisco_config_exception,
)
from petisco.events.subscriber.domain.config_event_subscriber import (
    ConfigEventSubscriber,
)
from petisco.events.subscriber.infrastructure.not_implemented_event_subscriber import (
    NotImplementedEventSubscriber,
)


@dataclass
class ConfigEventsSubscriber:
    provider: Optional[Callable] = lambda subscribers: NotImplementedEventSubscriber(
        subscribers
    )
    subscribers: Optional[Dict[str, ConfigEventSubscriber]] = None

    @staticmethod
    def from_dict(kdict):

        provider = kdict.get("provider")
        if not provider:
            raise TypeError(
                f"ConfigEventsSubscriber: {provider} is a required parameter"
            )
        provider = (
            get_function_from_string(provider)
            .handle(
                on_failure=raise_petisco_config_exception,
                failure_args=(kdict, "events:subscriber:provider"),
            )
            .unwrap()
        )

        subscribers = {}
        subscribers_dict = kdict.get("subscribers")
        if subscribers_dict:
            if not isinstance(subscribers_dict, dict):
                raise TypeError(
                    f"ConfigEventManager: subscribers must be a Dict with information about event subscribers"
                )
            subscribers = {
                k: ConfigEventSubscriber.from_dict(v)
                for k, v in subscribers_dict.items()
            }

        return ConfigEventsSubscriber(provider=provider, subscribers=subscribers)



from typing import Optional, Callable

from dataclasses import dataclass

from petisco.application.config.get_funtion_from_string import get_function_from_string
from petisco.application.config.raise_petisco_config_error import (
    raise_petisco_config_exception,
)


@dataclass
class ConfigProviders:
    config_dependencies: Optional[Callable] = None
    services_provider: Optional[Callable] = None
    repositories_provider: Optional[Callable] = None

    @staticmethod
    def from_dict(kdict):
        config_dependencies = (
            get_function_from_string(kdict.get("config_dependencies"))
            .handle(
                on_failure=raise_petisco_config_exception,
                failure_args=(kdict, "providers:config_dependencies"),
            )
            .unwrap()
        )
        services_provider = (
            get_function_from_string(kdict.get("services_provider"))
            .handle(
                on_failure=raise_petisco_config_exception,
                failure_args=(kdict, "providers:services_provider"),
            )
            .unwrap()
        )
        repositories_provider = (
            get_function_from_string(kdict.get("repositories_provider"))
            .handle(
                on_failure=raise_petisco_config_exception,
                failure_args=(kdict, "providers:repositories_provider"),
            )
            .unwrap()
        )

        return ConfigProviders(
            config_dependencies=config_dependencies,
            services_provider=services_provider,
            repositories_provider=repositories_provider,
        )



from typing import List, Any

from meiga import Result
from meiga.decorators import meiga

from petisco.logger.interface_logger import ERROR, INFO, DEBUG
from petisco.logger.log_message import LogMessage
from petisco.logger.not_implemented_logger import NotImplementedLogger
from petisco.use_case.use_case import UseCase


class _UseCaseHandler:
    def __init__(
        self,
        logger=NotImplementedLogger(),
        logging_parameters_whitelist: List[str] = None,
        logging_types_blacklist: List[Any] = None,
    ):
        self.logger = logger
        self.logging_parameters_whitelist = logging_parameters_whitelist
        self.logging_types_blacklist = logging_types_blacklist

    def __call__(self, cls):
        if not issubclass(cls, UseCase):
            raise TypeError("@use_case_handler only decorates UseCase subclasses")

        class UseCaseWrapped(cls):
            logger = self.logger
            logging_parameters_whitelist = self.logging_parameters_whitelist
            logging_types_blacklist = self.logging_types_blacklist

            def execute(self, *args, **kwargs):
                info_id = kwargs.get("info_id")

                log_message = LogMessage(
                    layer="use_case", operation=f"{cls.__name__}", info_id=info_id
                )

                self.logger.log(INFO, log_message.set_message(f"Running Use Case"))

                if self.logging_parameters_whitelist:
                    loggable_kwargs = {}
                    for key, value in kwargs.items():
                        if key not in self.logging_parameters_whitelist:
                            continue
                        if getattr(value, "to_json", None):
                            loggable_kwargs[key] = value.to_json()
                        else:
                            loggable_kwargs[key] = value

                    if loggable_kwargs:
                        self.logger.log(DEBUG, log_message.set_message(loggable_kwargs))

                result = self._run_execute(*args, **kwargs)

                if not isinstance(result, Result):
                    return result

                detail = getattr(result.value, "message", None)
                if detail:
                    detail = f"-> {detail}"
                else:
                    detail = ""

                if result.is_failure:
                    self.logger.log(
                        ERROR, log_message.set_message(f"{result} {detail}")
                    )
                else:
                    if not self._is_logging_type(result.value):
                        message = (
                            f"Success result of type: {type(result.value).__name__}"
                        )
                    else:
                        message = f"{result.value}"
                    self.logger.log(DEBUG, log_message.set_message(message))

                return result

            def _is_logging_type(self, value):
                if self.logging_types_blacklist:
                    for logging_type in self.logging_types_blacklist:
                        if isinstance(value, logging_type):
                            return False
                return True

            @meiga
            def _run_execute(self, *args, **kwargs) -> Result:
                return super().execute(*args, **kwargs)

        return UseCaseWrapped


use_case_handler = _UseCaseHandler



import os

from setuptools import setup

CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))
PACKAGE_NAME = "petisco"
VERSION = open("petisco/VERSION", "r").read()

# The text of the README file
with open(os.path.join(CURRENT_DIR, "README.md")) as fid:
    README = fid.read()

with open("requirements/requirements.txt") as f:
    required = f.read().splitlines()

setup(
    name=PACKAGE_NAME,
    version=VERSION,
    description="Petisco is a framework for helping Python developers to build clean Applications",
    long_description=README,
    long_description_content_type="text/markdown",
    keywords=["DDD", "Use Case", "Clean Architecture", "REST", "Applications"],
    url="https://github.com/alice-biometrics/petisco",
    author="ALiCE Biometrics",
    author_email="support@alicebiometrics.com",
    license="MIT",
    install_requires=required,
    entry_points={
        "console_scripts": ["petisco = petisco.application.cli.petisco:main"]
    },
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
    ],
    packages=[
        "petisco",
        "petisco/application",
        "petisco/application/config",
        "petisco/application/config/events",
        "petisco/application/config/cron",
        "petisco/application/cli",
        "petisco/controller",
        "petisco/controller/errors",
        "petisco/domain",
        "petisco/domain/value_objects",
        "petisco/domain/aggregate_roots",
        "petisco/domain/errors",
        "petisco/events",
        "petisco/events/rabbitmq",
        "petisco/events/publisher",
        "petisco/events/publisher/domain",
        "petisco/events/publisher/infrastructure",
        "petisco/events/subscriber",
        "petisco/events/subscriber/domain",
        "petisco/events/subscriber/infrastructure",
        "petisco/cron",
        "petisco/cron/domain",
        "petisco/cron/infrastructure",
        "petisco/fixtures",
        "petisco/frameworks",
        "petisco/frameworks/flask",
        "petisco/frameworks/flask/application",
        "petisco/logger",
        "petisco/persistence",
        "petisco/persistence/sqlalchemy",
        "petisco/security/token_decoder",
        "petisco/security/token_manager",
        "petisco/tools",
        "petisco/use_case",
        "petisco/modules",
        "petisco/modules/environment",
        "petisco/modules/environment/application",
        "petisco/modules/environment/domain",
        "petisco/modules/healthcheck",
        "petisco/modules/healthcheck/application",
        "petisco/modules/healthcheck/domain",
    ],
    include_package_data=True,
    zip_safe=False,
    extras_require={
        "flask": ["connexion==2.6.0", "connexion[swagger-ui]", "Flask-Cors==3.0.7"],
        "sqlalchemy": ["sqlalchemy>=1.3.11", "sqlalchemy_utils", "PyMySQL==0.9.2"],
        "redis": ["redis >= 3.3.11", "fakeredis>=1.0.5"],
        "rabbitmq": ["pika==1.1.0"],
        "gunicorn": ["gunicorn", "json-logging-py==0.2"],
        "fixtures": ["pytest"],
    },
)



from typing import Dict

from petisco.events.subscriber.domain.interface_event_subscriber import IEventSubscriber


class NotImplementedEventSubscriber(IEventSubscriber):
    def info(self) -> Dict:
        return {"name": self.__class__.__name__}

    def subscribe_all(self):
        pass

    def unsubscribe_all(self):
        pass



import pytest

from petisco.application.petisco import Petisco


@pytest.fixture
def petisco_client(petisco_sql_database):
    app = Petisco.get_instance().get_app()
    with app.app.test_client() as c:
        yield c



from abc import ABCMeta, abstractmethod

from petisco.application.config.cron.config_cron import ConfigCron


class CronExecutor:

    __metaclass__ = ABCMeta

    @abstractmethod
    def start(self, config_cron: ConfigCron):
        raise NotImplementedError

    @abstractmethod
    def stop(self):
        raise NotImplementedError



import re
from datetime import datetime
from typing import Dict

from petisco.events.event_id import EventId, EVENT_ID_LENGTH

import json

from typing import List

TIME_FORMAT = "%Y-%m-%d %H:%M:%S.%f"


class Event:
    event_id: EventId = None
    event_name: str = None
    event_occurred_on: str = None
    event_version: int = None
    event_info_id: Dict = None

    def __init__(self, dictionary=None):

        self.event_version = 1

        if dictionary:
            self.__dict__.update(dictionary)
            self.event_version = dictionary.get("event_version", 1)

        self._set_id(dictionary)
        self._set_name()
        self._set_occurred_on()

    def _set_id(self, dictionary: dict):
        self.event_id = (
            EventId.generate(str(dictionary)) if not self.event_id else self.event_id
        )

    def _set_name(self):
        self.event_name = (
            self.__class__.__name__ if not self.event_name else self.event_name
        )
        self.event_name = (
            re.sub(r"(?<!^)(?=[A-Z])", "_", self.event_name).lower().replace("_", ".")
        )

    def _set_occurred_on(self):
        self.event_occurred_on = (
            datetime.utcnow() if not self.event_occurred_on else self.event_occurred_on
        )

    def __repr__(self):
        return f"{json.loads(self.to_json())}"

    def __eq__(self, other):
        if (
            isinstance(other, self.__class__)
            or issubclass(other.__class__, self.__class__)
            or issubclass(self.__class__, other.__class__)
        ):
            return self.to_dict() == other.to_dict()
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def to_dict(self) -> Dict:

        raw_dict = self.__dict__.copy()
        data = {
            "data": {
                "id": str(raw_dict.pop("event_id")),
                "type": raw_dict.pop("event_name"),
                "version": str(raw_dict.pop("event_version")),
                "occurred_on": raw_dict.pop("event_occurred_on").strftime(TIME_FORMAT),
                "attributes": {},
                "meta": {},
            }
        }

        if "event_info_id" in raw_dict:
            if raw_dict.get("event_info_id"):
                data["data"]["meta"]["info_id"] = raw_dict.get("event_info_id")
            raw_dict.pop("event_info_id")

        data["data"]["attributes"] = raw_dict

        return data

    @staticmethod
    def from_dict(jsonapi: Dict):

        data = jsonapi.get("data")

        event_dictionary = {}

        if "id" in data and isinstance(data["id"], str):
            event_dictionary["event_id"] = EventId(data["id"])

        if "type" in data and isinstance(data["type"], str):
            event_dictionary["event_name"] = EventId(data["type"])

        if "version" in data and isinstance(data["version"], str):
            event_dictionary["event_version"] = int(data["version"])

        if "occurred_on" in data and isinstance(data["occurred_on"], str):
            event_dictionary["event_occurred_on"] = datetime.strptime(
                data["occurred_on"], TIME_FORMAT
            )

        event_dictionary["event_info_id"] = data.get("meta", {}).get("info_id")

        event_dictionary.update(data.get("attributes"))

        return Event(event_dictionary)

    @staticmethod
    def from_deprecated_dict(deprecated_dict: Dict):
        if "event_id" in deprecated_dict and isinstance(
            deprecated_dict["event_id"], str
        ):
            new_event_id = deprecated_dict["event_id"].rjust(EVENT_ID_LENGTH, "0")
            deprecated_dict["event_id"] = EventId(new_event_id)
        if "event_occurred_on" in deprecated_dict and isinstance(
            deprecated_dict["event_occurred_on"], str
        ):
            deprecated_dict["event_occurred_on"] = datetime.strptime(
                deprecated_dict["event_occurred_on"], TIME_FORMAT
            )

        if (
            "event_version" in deprecated_dict
            and isinstance(deprecated_dict["event_version"], str)
            and len(deprecated_dict["event_version"]) > 1
        ):
            deprecated_dict["event_version"] = "0"

        info_id = None
        if "info_id" in deprecated_dict and isinstance(
            deprecated_dict["info_id"], dict
        ):
            from petisco.domain.aggregate_roots.info_id import InfoId

            info_id = InfoId.from_dict(deprecated_dict.pop("info_id"))

        event = Event(deprecated_dict)
        event.add_info_id(info_id)

        return event

    def to_json(self):
        return json.dumps(self.to_dict(), default=self._datetime_to_str)

    def _datetime_to_str(self, o):
        if isinstance(o, datetime):
            return o.__str__()

    @staticmethod
    def from_json(event_json):
        event_dict = json.loads(event_json)
        return Event.from_dict(event_dict)

    @staticmethod
    def from_deprecated_json(event_json):
        event_dict = json.loads(event_json)
        return Event.from_deprecated_dict(event_dict)

    def add_info_id(self, info_id):
        if not info_id:
            return self

        from petisco.domain.aggregate_roots.info_id import (
            InfoId,
        )  # internal import to avoid circular dependency

        if not isinstance(info_id, InfoId):
            raise TypeError("Event.add_info_id() expect a InfoId class (from petisco)")
        self.event_info_id = info_id.to_dict()
        return self


Events = List[Event]
